/*
 * Extension d'ImageView grant les zooms
 * Copyright - Fars Belhadj - 2013
 * email: amsi@ai.univ-paris8.fr
 */
package fr.fbelhadj.imgzoom;
import android.annotation.SuppressLint;
import android.content.Context;
import android.graphics.Bitmap;
import android.graphics.Canvas;
import android.graphics.Matrix;
import android.graphics.Rect;
import android.util.AttributeSet;
import android.util.FloatMath;
import android.view.GestureDetector;
import android.view.MotionEvent;
import android.view.View;
import android.widget.ImageView;
@SuppressLint("FloatMath")
public class MyImageView extends ImageView {
	enum MouseStates {
		NONE,
		DRAG,
		ZOOM
	};
	private int _dw = 1, _dh = 1;
	private ImageConverger _ic = new ImageConverger(this);
	private MouseStates _ms = MouseStates.NONE;
	private float _ratio = 1, _space = 0, _zf = 1, _zf0 = 1, _tx = 0, _ty = 0;
	private Matrix _curMatrix = new Matrix(), _oldMatrix = new Matrix(), _applyedMatrix = new Matrix();
	private Rect _bounds = new Rect();
	private boolean _disableDbleTap = false;	
	public MyImageView(Context context, AttributeSet attrs, int defStyle) {
		super(context, attrs, defStyle);
		init();
	}
	public MyImageView(Context context, AttributeSet attrs) {
		super(context, attrs);
		init();
	}
	public MyImageView(Context context) {
		super(context);
		init();
	}
	private void init() {
		setScaleType(ImageView.ScaleType.MATRIX);
		setOnTouchListener(new OnTouchListener() {
			@Override
			public boolean onTouch(View v, MotionEvent event) {
				return manageOnTouch(v, event);
			}
		});
	}
	private float[] computeImageLimits(Matrix mat) {
		float t[] = {_dw, _dh}, l[] = { 0, 0, 0, 0 };
		float w = _bounds.right - _bounds.left;
		float h = _bounds.bottom - _bounds.top;
		mat.mapVectors(t);
		t[0] = Math.max(0, (w - t[0]) / 2);
		t[1] = Math.max(0, (h - t[1]) / 2);
		l[0] = t[0];
		l[1] = t[1];
		l[2] = w - t[0];
		l[3] = h - t[1];
		return l;
	}	
	private void applyImageMatrix() {
		_applyedMatrix = new Matrix(_curMatrix);
		setImageMatrix(_applyedMatrix);
	}
	private void updateMatrix() {
		float w = _bounds.width(), h = _bounds.height();
		_ratio = w / _dw;
		if(_dh > h / _ratio) _ratio = h / _dh;
		float limits[];
    	_zf0 = _zf = 1f;
		_space = _tx = _ty = 0f;
		_curMatrix.setScale(_ratio, _ratio);
		limits = computeImageLimits(_curMatrix);
		_curMatrix.postTranslate(limits[0], limits[1]);
		_oldMatrix.set(_curMatrix);
		applyImageMatrix();
	}
    private void fixZoomSpace(boolean twoCoords, MotionEvent event) {
    	if(twoCoords) {
        	_space = FloatMath.sqrt( (event.getX(0) - event.getX(1)) * (event.getX(0) - event.getX(1)) + 
        			(event.getY(0) - event.getY(1)) * (event.getY(0) - event.getY(1)) );
    		_tx = (event.getX(0) + event.getX(1)) / 2;
    		_ty = (event.getY(0) + event.getY(1)) / 2;
    	} else {
    		_space = 0f;
    		_tx = event.getX(0); 
    		_ty = event.getY(0);
    	}
    	_zf0 /= _zf; 
    	_zf = 1.0f;
    }
    private boolean updateZoom(MotionEvent event) {
    	float ns = FloatMath.sqrt( (event.getX(0) - event.getX(1)) * (event.getX(0) - event.getX(1)) + 
    			(event.getY(0) - event.getY(1)) * (event.getY(0) - event.getY(1)) );
		_tx = (event.getX(0) + event.getX(1)) / 2;
		_ty = (event.getY(0) + event.getY(1)) / 2;
    	if(Float.isNaN(ns) || Float.isNaN(_space) || _space < 10 || ns < 10)
    		return false;
		_zf = ns / _space;
        _zf = _zf < _zf0 ? _zf0 : _zf;
        return true;
    }
	@SuppressLint("DrawAllocation")
	@Override
	protected void onLayout(boolean changed, int left, int top, int right, int bottom) {
		_bounds = new Rect(left, top, right, bottom);
		updateMatrix();
	}
    private void recenter(MotionEvent event, boolean twoCoords, boolean reinit) {
    	_ic.setShouldConverge(false);
    	if(reinit) {
    		fixZoomSpace(twoCoords, event);
            _oldMatrix.set(_curMatrix);
    	} else if(twoCoords) {
        	if(!updateZoom(event))
        		return;
    	}
        if((event.getAction() & MotionEvent.ACTION_MASK) == MotionEvent.ACTION_MOVE) {
        	if (twoCoords) {
				_curMatrix.set(_oldMatrix);
				_curMatrix.postScale(_zf, _zf, _tx, _ty);
        	} else {
	    		Matrix tmp = new Matrix(_curMatrix);
				float pts[] = {0, 0, _dw, _dh}; 
				float dx = event.getX() - _tx, dy = event.getY() - _ty;
				tmp.postTranslate(dx, dy);
				tmp.mapPoints(pts);
				_curMatrix.postTranslate(dx, dy);
				_tx += dx; _ty += dy;
			}
        } else
        	_oldMatrix.set(_curMatrix);
        applyImageMatrix();
    }
	@Override
	public void draw(Canvas canvas) {
		super.draw(canvas);
		_ic.loop();
	}
	private Matrix _curMatrixSave = new Matrix();
	@SuppressWarnings("deprecation")
	private GestureDetector _gd = new GestureDetector(new GestureDetector.SimpleOnGestureListener() {
		@Override
		public boolean onDoubleTap(MotionEvent e) {
			if(_disableDbleTap) return false;
			_curMatrix.set(_oldMatrix);
			_zf *= 2;
			_curMatrix.postScale(_zf, _zf, _tx = e.getX(), _ty = e.getY());
			applyImageMatrix();
			return true;
		}
	});
    public boolean manageOnTouch(View view, MotionEvent event) {
    	if(!_disableDbleTap && _gd.onTouchEvent(event)) {
    		_ic.setShouldConverge(true);
    		return true;
    	}
    	if(_ic._shouldConverge) return false;
    	switch (event.getAction() & MotionEvent.ACTION_MASK) {
    	case MotionEvent.ACTION_DOWN:
    		_ms = MouseStates.DRAG;
    		recenter(event, false, true);
    		_curMatrixSave = new Matrix(_curMatrix);
    		break;
    	case MotionEvent.ACTION_POINTER_DOWN:
    		_ms = MouseStates.ZOOM;
    		recenter(event, true, true);
    		break;
    	case MotionEvent.ACTION_UP:
    	case MotionEvent.ACTION_POINTER_UP:
    		if(_ms == MouseStates.NONE)
    			break;
    		_ms = MouseStates.NONE;
    		if(!_curMatrixSave.equals(_curMatrix))
    			_ic.setShouldConverge(true);
    		break;
    	case MotionEvent.ACTION_MOVE:
    		if (_ms == MouseStates.ZOOM || (_ms == MouseStates.DRAG && isScaled()))
    			recenter(event, _ms == MouseStates.ZOOM, false);
    		break;
    	}
    	return true;
    }
    @Override
	public void setImageBitmap(Bitmap bm) {
    	super.setImageBitmap(bm);
    	if(bm != null) {
    		_dw = bm.getWidth();
    		_dh = bm.getHeight();
    	}
    	updateMatrix();
    }
    static boolean approx(float a, float b) {
    	return Math.abs(a - b) < 0.0001;
    }
	public boolean isScaled() {
		float values[] = new float[9];
		_curMatrix.getValues(values);
		return (!approx(values[0], _ratio)); 
	}
	public void setDisableDoubleTap(boolean disable) {
		_disableDbleTap = disable;
	}
	private class ImageConverger {
		private View _parent;
		private boolean _shouldConverge = false,  _shouldConverge0 = false;
		private float _converge_dx = 0, _converge_dy = 0, _converge_count = 0, _converge_count_max = 0;
		private long _t0;
		private void initTime() {
			_t0 = System.currentTimeMillis();
		}
		private long getTime() {
			return _t0 = (System.currentTimeMillis() - _t0);
		}
		public ImageConverger(View parent) {
			_parent = parent;
		}
		public void setShouldConverge(boolean shouldConverge) {
			if(shouldConverge) {
				initTime();
				_parent.invalidate();
			}
			_shouldConverge = _shouldConverge0 = shouldConverge;
		}
		private void start() {
			float pts[] = {0, 0, _dw, _dh}, limits[];
			_converge_dx = _converge_dy = 0;
			_curMatrix.mapPoints(pts);
			limits = computeImageLimits(_curMatrix);
			if(pts[0] > limits[0]) _converge_dx = limits[0] - pts[0];
			if(pts[2] < limits[2]) _converge_dx = limits[2] - pts[2];
			if(pts[1] > limits[1]) _converge_dy = limits[1] - pts[1];
			if(pts[3] < limits[3]) _converge_dy = limits[3] - pts[3];
			if(!(_converge_dx < 1.0 && _converge_dx > -1.0) || !(_converge_dy < 1.0 && _converge_dy > -1.0))
				_converge_count = Math.max(Math.min(1, 1000f / getTime()), _converge_count_max = 6);
			else
				_converge_count = 0;
			_shouldConverge0 = false;
		}
	    public void loop() {
	    	if(_shouldConverge0)
	    		start();
	    	if(_shouldConverge) { 
				if(_converge_count > 0) {
					float dx = _converge_dx / _converge_count, dy = _converge_dy / _converge_count;
					_converge_dx -= dx; _converge_dy -= dy; 
					_curMatrix.postTranslate(dx, dy);
					applyImageMatrix();
					_shouldConverge = true;
					if(--_converge_count > 0)
						_converge_count = Math.max(Math.min(1, 1000f / getTime()), --_converge_count_max);
				} else {
					_oldMatrix.set(_curMatrix);
					_shouldConverge = false;
				}
	    	}
	    }
	}
}
